import subprocess
import shlex
import os

from rlpython.utils.argument_parser import (
    ReplArgumentParserError,
    ReplArgumentParser,
)


class ShellRuntime:
    def __init__(self, repl):
        self.repl = repl
        self._old_pwd = os.getcwd()

    def validate_source(self, raw_source):
        try:
            shlex.split(raw_source)

            return True

        except Exception:
            return False

    def change_directory(self, command_line):
        argument_parser = ReplArgumentParser(repl=self.repl, prog='cd')
        argument_parser.add_argument('directory', nargs='?')

        arguments = argument_parser.parse_args(command_line[1:])
        directory = arguments.directory

        if directory == '-':
            directory = self._old_pwd

        else:
            if not directory:
                directory = os.environ.get('HOME', '/home')

            directory = os.path.abspath(directory)

        self._old_pwd = os.getcwd()

        try:
            os.chdir(directory)

        except FileNotFoundError:
            self.repl.write_error(
                'cd: {}: No such file or directory\n'.format(directory),
            )

            return 1

        self.repl.write('{}\n'.format(directory))

        return 0

    def run(self, raw_source):
        # split into command line
        source = raw_source[1:]
        command_line = shlex.split(source)

        # change_directory command
        if command_line[0] == 'cd':
            try:
                return self.change_directory(command_line)

            except ReplArgumentParserError:
                return 1

        # run command
        try:
            output = subprocess.check_output(
                command_line,
                stderr=subprocess.PIPE,
                shell=False,
            ).decode()

            exit_code = 0

        except FileNotFoundError:
            output = ''
            exit_code = 1

            self.repl.write_error(
                '{}: command not found\n'.format(command_line[0]),
            )

        except subprocess.CalledProcessError as exception:
            output = '{}{}'.format(
                exception.output.decode(),
                exception.stderr.decode(),
            )

            exit_code = exception.returncode

        if output:
            self.repl.write(output)

        return exit_code



from textwrap import wrap

from rlpython.utils.asyncio_utils import get_all_loops, get_all_tasks
from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.gc_utils import get_object_by_id
from rlpython.utils.table import write_table
from rlpython.utils.strings import color


class TasksCommand:
    """
    List all running asyncio tasks
    """

    NAME = 'tasks'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='tasks')
        argument_parser.add_argument('loop-id', nargs='?')

        arguments = vars(argument_parser.parse_args(argv[1:]))

        # find loops
        if arguments['loop-id']:
            loop = get_object_by_id(arguments['loop-id'])

            if not loop:
                self.repl.write(color('ERROR: invalid loop id\n', fg='red'))

                return 1

            loops = [loop]

        else:
            loops = get_all_loops()

        loops = sorted(loops, key=lambda i: id(i))

        # write table
        rows = [
            ['Loop ID', 'Task ID', 'State', 'Task'],
        ]

        for loop in loops:
            tasks = sorted(get_all_tasks(loop=loop), key=lambda i: id(i))

            for task in tasks:
                rows.append([
                    str(id(loop)),
                    str(id(task)),
                    str(task._state),
                    '\n'.join(wrap(repr(task), 80)),
                ])

        if len(rows) == 1:
            return

        write_table(rows, self.repl.write)



from rlpython.protocol import encode_edit_message

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.editor import NoEditorError, run_editor
from rlpython.utils.strings import color


class EditCommand:
    """
    Open source code of given object in $EDITOR if source code is available
    """

    NAME = 'edit'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='edit')
        argument_parser.add_argument('object')

        arguments = argument_parser.parse_args(argv[1:])

        # find object
        obj = self.repl.python_runtime.eval(arguments.object)

        if not obj:
            self.repl.write(
                color("ERROR: '{}' not found\n".format(arguments.object), fg='red'),  # NOQA
            )

            return 1

        # find file
        filename, lineno = self.repl.python_runtime.get_file(obj)

        if not filename:
            self.repl.write(
                color("ERROR: '{}' source file not found\n".format(repr(obj)), fg='red'),  # NOQA
            )

            return 1

        if not filename.endswith('.py'):
            self.repl.write(
                color("ERROR: '{}' is no python file\n".format(filename), fg='red'),  # NOQA
            )

            return 1

        # local
        if self.repl.domain == self.repl.DOMAIN.LOCAL:
            try:
                run_editor(filename, lineno=lineno)

            except NoEditorError:
                self.repl.write(
                    color('ERROR: no editor found\n', fg='red'),
                )

        # network
        else:
            encode_text = False

            if self.repl.domain == self.repl.DOMAIN.NETWORK:
                encode_text = True

            message = encode_edit_message(
                filename=filename,
                lineno=lineno,
                encode_text=encode_text,
            )

            self.repl.write_message(message)



from argparse import ArgumentParser


class ReplArgumentParserError(Exception):
    pass


class ReplArgumentParser(ArgumentParser):
    def __init__(self, repl, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._repl = repl

    def _print_message(self, message, file=None):
        self._repl.write(message)

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message)

        raise ReplArgumentParserError



import logging
import os


class SilentLogger:
    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def _debug_enabled(self):
        value = os.environ.get('RLPYTHON_DEBUG', '').strip().lower()

        return value in ('1', 'enabled', 'enable', 'true')

    def debug(self, *args, **kwargs):
        if not self._debug_enabled():
            return

        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.logger.critical(*args, **kwargs)



import threading
import traceback
import sys

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table
from rlpython.utils.strings import color


class ThreadsCommand:
    """
    List all running threads; Show stack of given thread
    """

    NAME = 'threads'

    def __init__(self, repl):
        self.repl = repl

    def complete(self, text, state, line_buffer):
        names = []

        for thread in threading.enumerate():
            names.append(str(thread.getName()))

        candidates = []

        for name in sorted(names):
            if name.startswith(text):
                candidates.append(name)

        candidates.append(None)

        return candidates[state]

    def find_frame(self, identifier):
        for thread in threading.enumerate():
            if(str(thread.ident) == identifier or
               thread.getName() == identifier):

                return thread, sys._current_frames()[thread.ident]

        return None, None

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='threads')
        argument_parser.add_argument('identifier', nargs='?')

        arguments = vars(argument_parser.parse_args(argv[1:]))

        # show stack of given thread
        if arguments['identifier']:
            thread, frame = self.find_frame(arguments['identifier'])

            if not thread:
                self.repl.write(
                    color('ERROR: invalid thread ident\n', fg='red')
                )

                return 1

            self.repl.write('STACK: {} id={}\n'.format(
                thread.getName(),
                str(thread.ident),
            ))

            for line in traceback.format_stack(frame):
                self.repl.write(line)

            self.repl.write('END STACK\n')

            return 0

        # write table
        rows = [['Thread ID', 'Thread Name', 'Alive', 'Daemon', 'Task']]

        current_frames = sys._current_frames()

        for thread in threading.enumerate():
            frame = current_frames[thread.ident]

            task = '{}:{} {}'.format(
                frame.f_code.co_filename,
                frame.f_lineno,
                frame.f_code.co_name,
            )

            rows.append([
                thread.ident,
                thread.getName(),
                thread.isAlive(),
                thread.isDaemon(),
                task,
            ])

        write_table(rows, self.repl.write)



from argparse import ArgumentParser
import signal
import os


def handle_command_line():
    from rlpython.repl_client import ReplClient
    from rlpython import embed

    argument_parser = ArgumentParser(prog='rlpython')

    argument_parser.add_argument('url', nargs='?', type=str)
    argument_parser.add_argument('--history-file', type=str)
    argument_parser.add_argument('--history-size', type=int)
    argument_parser.add_argument('--banner', type=str)
    argument_parser.add_argument('--prompt', type=str)
    argument_parser.add_argument('--prompt-ps2', type=str)
    argument_parser.add_argument('--frontend-mode', action='store_true')

    namespace = argument_parser.parse_args()

    # frontend mode
    if namespace.frontend_mode:
        def restore_pgrp(*args, **kwargs):
            os.tcsetpgrp(fd, old_pgrp)
            os.setpgid(0, old_pgrp)

            exit(0)

        signal.signal(signal.SIGTSTP, signal.SIG_IGN)
        signal.signal(signal.SIGTTOU, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, restore_pgrp)

        pid = os.getpid()
        fd = os.open('/dev/tty', os.O_RDONLY)
        old_pgrp = os.tcgetpgrp(fd)

        os.setpgid(0, pid)
        os.tcsetpgrp(fd, pid)

    # client mode
    if namespace.url:
        try:
            repl_client = ReplClient(
                url=namespace.url,
                frontend_mode=namespace.frontend_mode,
            )

        except Exception as exception:
            exit(str(exception))

        try:
            repl_client.interact()

        finally:
            repl_client.shutdown()

            if namespace.frontend_mode:
                restore_pgrp()

        if namespace.frontend_mode:
            restore_pgrp()

        exit(0)

    # local mode
    repl_kwargs = {}

    for key in dir(namespace):
        if key.startswith('_'):
            continue

        if key in ('frontend_mode'):
            continue

        value = getattr(namespace, key)

        if value is None:
            continue

        repl_kwargs[key] = value

    try:
        embed(
            globals={},
            locals={},
            single_threaded=True,
            started_from_cmd_line=True,
            **repl_kwargs,
        )

    finally:
        if namespace.frontend_mode:
            restore_pgrp()


if __name__ == '__main__':
    handle_command_line()



from tempfile import NamedTemporaryFile
import readline
import socket
import time
import os

from rlpython.utils.editor import run_editor
from rlpython.utils.url import parse_url
from rlpython.repl import Repl

from rlpython.protocol import (
    encode_completion_request_message,
    END_OF_TRANSMISSION,
    encode_ping_message,
    encode_run_message,
    decode_message,
    MESSAGE_TYPE,
)


class ReplClient(Repl):
    def __init__(self, url, *args, frontend_mode=False, **kwargs):
        self.url = url

        self.scheme, self.host, self.port = parse_url(url)

        # setup prompt prefix
        if frontend_mode:
            prompt_prefix = ''

        else:
            if self.scheme == 'file':
                prompt_prefix = 'file://{} '.format(self.host)

            else:
                prompt_prefix = '{}:{} '.format(self.host, self.port)

        super().__init__(
            *args,
            prompt_prefix=prompt_prefix,
            **kwargs,
        )

    def send_message(self, message):
        exit_code, payload = message

        if not exit_code:
            return

        try:
            self.sock.send(payload)

        except BrokenPipeError:
            pass

    def recv_message(self):
        message_buffer = bytes()

        while True:
            try:
                data = self.sock.recv(1)

            except BrokenPipeError:
                return True, None, None

            if not data:
                return True, None, None

            message_buffer += data

            if message_buffer[-1] == END_OF_TRANSMISSION[0]:
                exit_code, message_type, payload = decode_message(
                    message_buffer,
                )

                if exit_code:
                    return False, message_type, payload

                message_buffer = bytes()

                continue

    def setup(self):
        # network
        if self.scheme == 'rlpython':
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            connect_arg = (self.host, self.port)

        # unix domain socket
        elif self.scheme == 'file':
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            connect_arg = self.host

        # wait for connection
        while True:
            try:
                self.sock.connect(connect_arg)

                break

            except ConnectionRefusedError:
                time.sleep(1)

        # handle setup messages from server
        while True:
            disconnected, message_type, payload = self.recv_message()

            if disconnected:
                exit(1)

            # set
            if message_type == MESSAGE_TYPE.SET:
                name, value = payload

                if name in ('banner', 'warnings', 'ps1', 'ps2'):
                    setattr(self, name, value)

            # ready
            if message_type == MESSAGE_TYPE.READY:
                break

        self.write_banner()
        self.write_warnings()

    def complete(self, text, state):
        line_buffer = readline.get_line_buffer()

        # handle starting tabs in multi line statements locally
        if not line_buffer.strip() and not text.strip():
            if state == 0:
                readline.insert_text('\t')
                readline.redisplay()

                return ''

            return None

        # remote completion
        self.send_message(
            encode_completion_request_message(text, state, line_buffer)
        )

        while True:
            disconnected, message_type, payload = self.recv_message()

            if disconnected:
                exit(1)

            # write
            if message_type == MESSAGE_TYPE.WRITE:
                self.write(payload)

            # completion response
            elif message_type == MESSAGE_TYPE.COMPLETION_RESPONSE:
                return payload

    def handle_empty_line(self):
        self.clear_line_buffer()

        # send ping to server and wait for pong
        self.send_message(encode_ping_message())

        while True:
            disconnected, message_type, payload = self.recv_message()

            if disconnected:
                exit(1)

            # write
            if message_type == MESSAGE_TYPE.WRITE:
                self.write(payload)

            # pong
            if message_type == MESSAGE_TYPE.PONG:
                return

    def edit(self, filename, lineno, text):
        if text is not None:
            temp_file = NamedTemporaryFile(suffix='.py')
            filename = temp_file.name

            with open(filename, 'w') as f:
                f.write(text)
                f.close()

            filename = temp_file.name

        run_editor(filename=filename, lineno=lineno)

    def run(self, command):
        self.send_message(encode_run_message(command))

        while True:
            disconnected, message_type, payload = self.recv_message()

            if disconnected:
                exit(1)

            # write
            if message_type == MESSAGE_TYPE.WRITE:
                self.write(payload)

            # edit
            elif message_type == MESSAGE_TYPE.EDIT:
                filename, lineno, text = payload

                self.edit(filename, lineno, text)

            # exit code
            elif message_type == MESSAGE_TYPE.EXIT_CODE:
                self.exit_code = payload

                return



from rlpython.utils.strings import color


def write_attribute_table(rows, write, header_fg=None):
    header_length = max([len(row[0]) for row in rows]) + 1

    for row in rows:
        header = row[0].rjust(header_length)

        if header_fg is not None:
            header = color(header, fg=header_fg)

        write('{}: {}\n'.format(header, row[1]))



import textwrap
import os

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table
from rlpython.utils.strings import color


class UnixEnvironmentCommand:
    """
    List Unix environmet variables; Print value of given environmet variable
    """

    NAME = 'env'

    def __init__(self, repl):
        self.repl = repl

    def complete(self, text, state, line_buffer):
        names = [*sorted(os.environ.keys())]
        candidates = []

        for name in names:
            if name.startswith(text):
                candidates.append(name)

        candidates.append(None)

        return candidates[state]

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='env')
        argument_parser.add_argument('name', nargs='?')

        arguments = vars(argument_parser.parse_args(argv[1:]))

        # name
        if arguments['name']:

            # set
            if '=' in arguments['name']:
                name, value = arguments['name'].split('=')

                try:
                    value = self.repl.python_runtime.eval(
                        source=repr(value),
                        safe=False,
                    )

                except Exception:
                    self.repl.write(
                        color(
                            "ERROR: '{}' is no valid python expression\n".format(value),  # NOQA
                            fg='red',
                        ),
                    )

                    return 1

                os.environ[name] = value

            # get
            else:
                self.repl.write(
                    repr(os.environ.get(arguments['name'], '')) + '\n')

            return

        # write table
        rows = [
            ['Key', 'Value'],
        ]

        for key in sorted(os.environ.keys()):
            value = '\n'.join(textwrap.wrap(os.environ[key], 80))

            rows.append([key, value])

        write_table(rows, self.repl.write)



from pprint import pformat
import builtins
import inspect

from rlpython.utils.attribute_table import write_attribute_table


class PythonRuntime:
    def __init__(self, repl, globals={}, locals={}):
        self.repl = repl
        self.globals = globals
        self.locals = locals

        # override print
        self.globals['print'] = self.print_function
        self.globals['_print'] = builtins.print

        # prepare special keywords
        self.locals['_'] = None
        self.locals['_rlpython'] = self.repl
        self.locals['_exception'] = None

    def shutdown(self):
        del self.globals['print']
        del self.globals['_print']

        del self.locals['_']
        del self.locals['_rlpython']
        del self.locals['_exception']

    def print_function(self, *strings, end='\n'):
        string = ' '.join([str(i) for i in strings]) + end

        return self.repl.write(string)

    # introspection ###########################################################
    def get_file(self, obj):
        filename = None
        lineno = None

        try:
            filename = inspect.getabsfile(obj)

        except Exception:
            pass

        if not filename and hasattr(obj, '__class__'):
            try:
                obj = obj.__class__
                filename = inspect.getabsfile(obj)

            except Exception:
                pass

        if filename:
            try:
                lineno = inspect.getsourcelines(obj)[1]

            except Exception:
                pass

        return filename, lineno

    def get_file_string(self, obj):
        filename, lineno = self.get_file(obj)
        string = ''

        if filename:
            string = filename

            if lineno:
                string += ':' + str(lineno)

        return string

    # buffer helper ###########################################################
    def write_representation(self, value):
        if self.repl.variables.get('pretty_print', True):
            try:
                self.repl.write(pformat(value) + '\n')

                return

            except Exception:
                pass

        self.repl.write(repr(value) + '\n')

    def write_short_description(self, value):
        rows = []

        rows.append(['id', hex(id(value))])
        rows.append(['type', repr(type(value))])

        # filename
        filename = self.get_file_string(value)

        if filename:
            rows.append(['file', filename])

        # signature
        if callable(value):
            try:
                signature = '{}{}'.format(
                    value.__name__,
                    inspect.signature(value),
                )

                rows.append(['signature', signature])

            except Exception:
                pass

        # write to repl
        write_attribute_table(rows, self.repl.write, header_fg='red')

    def write_long_description(self, value):
        documentation = inspect.getdoc(value)

        if not documentation:
            return

        self.repl.write(documentation + '\n')

    def write_description(self, value, length):
        if length == 1:
            self.write_short_description(value)

        elif length == 2:
            self.write_long_description(value)

    # source code helper ######################################################
    def parse_source(self, raw_source):
        multiline = len(raw_source.splitlines()) > 1
        terminated = raw_source.endswith('\n\n')
        describe = 0
        source = raw_source.strip()

        # long description
        if source[len(source)-2:] == '??':
            source = source[:-2]
            describe = 2

        # short description
        elif source[-1] == '?':
            source = source[:-1]
            describe = 1

        return source, multiline, terminated, describe

    def validate_source(self, raw_source):
        source, multiline, terminated, describe = self.parse_source(raw_source)

        if multiline and not terminated:
            return False

        try:
            compile(source, filename='<string>', mode='exec')

            return True

        except Exception:
            return False

    # code running ############################################################
    def eval(self, source, safe=True):
        try:
            return eval(source, self.globals, self.locals)

        except Exception:
            if not safe:
                raise

            return None

    def run(self, raw_source):
        exit_code = 0
        source, multiline, terminated, describe = self.parse_source(raw_source)

        # compile source
        try:
            run_code = eval

            code = compile(
                source=source,
                filename='<input>',
                mode='eval',
            )

        except SyntaxError:
            code = None

        if not code:
            try:
                run_code = exec

                code = compile(
                    source=source,
                    filename='<input>',
                    mode='exec',
                )

            except SyntaxError as exception:
                self.locals['_exception'] = exception
                self.repl.write_exception(exception)

                return 1

        # run code
        try:
            return_value = run_code(code, self.globals, self.locals)

        except KeyboardInterrupt:
            exit_code = 1

            self.repl.write('\n')

        except Exception as exception:
            exit_code = 1

            self.locals['_exception'] = exception
            self.repl.write_exception(exception)

        else:
            if run_code is eval:
                self.locals['_'] = return_value

                if describe:
                    self.write_description(return_value, describe)

                elif return_value is not None:
                    self.write_representation(return_value)

        return exit_code



from urllib.parse import urlparse, splitport


def parse_url(raw_url):
    try:
        if isinstance(raw_url, int):
            raw_url = str(raw_url)

        if raw_url.startswith('file://'):
            return 'file', raw_url[7:], None

        parse_result = urlparse(raw_url)

        if not parse_result.scheme:
            raw_url = 'rlpython://{}'.format(raw_url)
            parse_result = urlparse(raw_url)

        scheme = parse_result.scheme
        host = parse_result.netloc
        port = parse_result.port

        if port is None:
            port = host
            host = 'localhost'

        host = splitport(host)[0]

        return scheme, host, int(port)

    except Exception:
        raise ValueError(
            'invalid url format. valid formats: [HOST:]PORT, file://PATH',
        )



from rlpython.logging import SilentLogger
import json

logger = SilentLogger('rlpython.protocol')

END_OF_TRANSMISSION = bytes([4])
MESSAGE_TYPES = range(1, 11)


class MESSAGE_TYPE:
    PING = 1
    PONG = 2
    SET = 3
    READY = 4
    COMPLETION_REQUEST = 5
    COMPLETION_RESPONSE = 6
    RUN = 7
    WRITE = 8
    EXIT_CODE = 9
    EDIT = 10


def encode_message(raw_message):
    try:
        message = json.dumps(raw_message)

        return True, message.encode() + END_OF_TRANSMISSION

    except Exception:
        logger.error(
            'exception raised while encoding %s',
            repr(raw_message),
            exc_info=True,
        )

        return False, None


def decode_message(raw_message):
    try:
        if raw_message[-1] == END_OF_TRANSMISSION[0]:
            raw_message = raw_message[:-1]

        message_type, payload = json.loads(raw_message.decode())

    except Exception:
        logger.error(
            'exception raised while decoding %s',
            repr(raw_message),
            exc_info=True,
        )

        return False, None, None

    if message_type not in MESSAGE_TYPES:
        logger.error('invalid message type: %s', repr(message_type))

        return False, None, None

    return True, message_type, payload


def encode_ping_message():
    return encode_message([MESSAGE_TYPE.PING, None])


def encode_pong_message():
    return encode_message([MESSAGE_TYPE.PONG, None])


def encode_set_message(name, value):
    return encode_message([MESSAGE_TYPE.SET, [name, value]])


def encode_ready_message():
    return encode_message([MESSAGE_TYPE.READY, None])


def encode_completion_request_message(text, state, line_buffer):
    return encode_message(
        [MESSAGE_TYPE.COMPLETION_REQUEST, [text, state, line_buffer]],
    )


def encode_completion_response_message(return_value):
    return encode_message([MESSAGE_TYPE.COMPLETION_RESPONSE, return_value])


def encode_run_message(command):
    return encode_message([MESSAGE_TYPE.RUN, command])


def encode_write_message(string):
    return encode_message([MESSAGE_TYPE.WRITE, string])


def encode_keyboard_interrupt_message():
    return encode_message([MESSAGE_TYPE.EXIT_CODE, None])


def encode_exit_code_message(exit_code):
    return encode_message([MESSAGE_TYPE.EXIT_CODE, exit_code])


def encode_edit_message(filename, lineno, encode_text=False):
    text = None

    if encode_text:
        text = open(filename, 'r').read()

    return encode_message([MESSAGE_TYPE.EDIT, [filename, lineno, text]])



import threading

from rlpython.utils.asyncio_utils import get_all_loops, get_all_tasks
from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class LoopsCommand:
    """
    List all running asyncio Loops
    """

    NAME = 'loops'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='loops')

        argument_parser.parse_args(argv[1:])

        # write table
        rows = [
            ['Loop ID', 'Thread ID', 'Thread Name', 'Running', 'Task Count'],
        ]

        threads = list(threading.enumerate())

        for loop in get_all_loops():
            thread = ''

            for _thread in threads:
                if _thread.ident == loop._thread_id:
                    thread = _thread.getName()

            rows.append([
                repr(id(loop)),
                loop._thread_id,
                thread,
                loop.is_running(),
                len(get_all_tasks(loop=loop)),
            ])

        if len(rows) == 1:
            return

        write_table(rows, self.repl.write)



import subprocess
import sys


def start_frontend(port):
    process = subprocess.Popen([
        sys.executable,
        '-m',
        'rlpython.command_line',
        'localhost:{}'.format(port),
        '--frontend',
    ])

    return process



from rlcompleter import Completer as rlCompleter
from itertools import chain
from glob import glob
import os


class Namespace(dict):
    def __init__(self, locals, globals):
        self.locals = locals
        self.globals = globals

    def items(self):
        return chain(
            self.locals.items(),
            self.globals.items(),
        )

    def __getitem__(self, name):
        if name in self.locals:
            return self.locals[name]

        if name in self.globals:
            return self.globals[name]

        raise KeyError


class Completer:
    def __init__(self, repl):
        self.repl = repl

        self.rlcompleter = rlCompleter(
            namespace=Namespace(
                locals=self.repl.locals,
                globals=self.repl.globals,
            ),
        )

    def python_complete(self, text, state):
        return self.rlcompleter.complete(text, state)

    def command_complete(self, text, state, line_buffer):
        # run custom completion
        if ' ' in line_buffer:
            command_name = line_buffer.split(' ', 1)[0]

            if(command_name and
               command_name in self.repl.command_runtime.commands):

                command = self.repl.command_runtime.commands[command_name]

                if hasattr(command, 'complete'):
                    return command.complete(text, state, line_buffer)

        # complete from command_runtime.commands
        candidates = []

        for name in sorted(self.repl.command_runtime.commands.keys()):
            if name.startswith(text):
                candidates.append(name)

        candidates.append(None)

        return candidates[state]

    def file_complete(self, text, state, line_buffer):
        path = os.path.normpath(text)

        if os.path.isdir(path) and path != '/':
            path += '/'

        candidates = []

        for candidate in glob(path + '*'):
            candidate = os.path.normpath(candidate)

            if os.path.isdir(candidate):
                candidate += '/'

            if candidate not in candidates:
                candidates.append(candidate)

        candidates.sort()
        candidates.append(None)

        return candidates[state]

    def complete(self, text, state, line_buffer):
        try:
            line_buffer = self.repl.aliases.resolve(line_buffer + '\n')

            # commands
            if line_buffer.startswith('%'):
                line_buffer = line_buffer[1:]

                return self.command_complete(
                    text=text,
                    state=state,
                    line_buffer=line_buffer,
                )

            # shell
            elif line_buffer.startswith('!'):
                return self.file_complete(
                    text=text,
                    state=state,
                    line_buffer=line_buffer,
                )

            # python
            return self.python_complete(text, state)

        except Exception as e:
            self.repl.write('\n')
            self.repl.write_exception(e, prefix='rlpython: ')



import textwrap
import shlex

from rlpython.commands.unix_environment import UnixEnvironmentCommand
from rlpython.utils.argument_parser import ReplArgumentParserError
from rlpython.commands.variables import VariablesCommand
from rlpython.commands.aliases import AliasesCommand
from rlpython.commands.threads import ThreadsCommand
from rlpython.commands.tasks import TasksCommand
from rlpython.commands.loops import LoopsCommand
from rlpython.commands.edit import EditCommand

DEFAULT_COMMANDS = [
    UnixEnvironmentCommand,
    VariablesCommand,
    AliasesCommand,
    ThreadsCommand,
    TasksCommand,
    LoopsCommand,
    EditCommand,
]


class CommandRuntime:
    def __init__(self, repl):
        self.repl = repl
        self.commands = {}

        for command in DEFAULT_COMMANDS:
            self.install_command(command)

    def shutdown(self):
        pass

    def install_command(self, command_class, name=''):
        command = command_class(self.repl)
        name = name or command.NAME

        self.commands[name] = command

    def validate_source(self, raw_source):
        try:
            shlex.split(raw_source)

            return True

        except Exception:
            return False

    def run(self, raw_source):
        argv = shlex.split(raw_source[1:])
        name = argv[0]

        if name not in self.commands:
            self.repl.write("ERROR: unknown command '{}'\n".format(name))

            return

        try:
            exit_code = self.commands[name].run(argv)

            if exit_code is None:
                exit_code = 0

        except ReplArgumentParserError:
            exit_code = 1

            pass

        except Exception as exception:
            exit_code = 1

            self.repl.write_exception(exception)

        return exit_code

    def gen_help_text(self):
        text = ''

        for command_name in sorted(self.commands.keys()):
            command = self.commands[command_name]

            text += '%{}'.format(command_name)

            if command.__doc__:
                doc_string = ''.join(command.__doc__).strip()

                text += '\n'

                text += textwrap.indent(
                    '\n'.join(textwrap.wrap(doc_string, 76)),
                    '    ',
                )

                text += '\n'

            text += '\n'

        return text



from copy import copy

DEFAULT_ALIASES = {
    'l': '!ls -hlF --color',
    'll': '!ls -hAlF --color',
    'pwd': '!pwd',
    'cd': '!cd',
    '..': '!cd ..',
    'touch': '!touch',
    'mkdir': '!mkdir',
    'cat': '!cat',
    'rm': '!rm',
}


class Aliases:
    def __init__(self, repl):
        self.repl = repl

        self.aliases = copy(DEFAULT_ALIASES)

    def resolve(self, command):
        # check for name collisions with repl variables
        if '=' in command:
            return command

        _command = command.strip()

        if(_command in self.repl.globals or
           _command in self.repl.locals):

            return command

        # resolve aliases
        for name, value in self.aliases.items():
            if not command.startswith(name):
                continue

            if(len(command) >= len(name) and
               command[len(name)] in (' ', '\n', '\t')):

                return '{}{}'.format(value, command[len(name):])

        return command



from pprint import pformat

from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table
from rlpython.utils.strings import color


class VariablesCommand:
    """
    List, read and write REPL variables
    """

    NAME = 'var'

    def __init__(self, repl):
        self.repl = repl

    def complete(self, text, state, line_buffer):
        names = []
        candidates = []

        for key in sorted(self.repl.variables.keys()):
            value = self.repl.variables[key]

            names.append('{}={}'.format(key, repr(value)))

        for name in names:
            if name.startswith(text):
                candidates.append(name)

        candidates.append(None)

        return candidates[state]

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='vars')
        argument_parser.add_argument('variable', nargs='?')

        arguments = argument_parser.parse_args(argv[1:])

        # print all variables
        if not arguments.variable:
            table = [
                ['Name', 'Value'],
            ]

            for key, value in self.repl.variables.items():
                table.append([
                    str(key),
                    pformat(value),
                ])

            write_table(table, self.repl.write)

            return 0

        # print variable
        if '=' not in arguments.variable:
            self.repl.write(
                repr(self.repl.variables.get(arguments.variable, '')) + '\n')

            return 0

        # set variable
        name, value = arguments.variable.split('=')

        try:
            value = self.repl.python_runtime.eval(value, safe=False)

        except Exception:
            self.repl.write(
                color(
                    "ERROR: '{}' is no valid python expression\n".format(value),  # NOQA
                    fg='red',
                ),
            )

            return 1

        self.repl.variables[name] = value



from rlpython.logging import SilentLogger
from rlpython.repl import Repl

from rlpython.protocol import (
    encode_completion_response_message,
    encode_exit_code_message,
    encode_ready_message,
    encode_write_message,
    encode_pong_message,
    END_OF_TRANSMISSION,
    encode_set_message,
    decode_message,
    MESSAGE_TYPE,
)

logger = SilentLogger('rlpython.server')


class SocketRepl(Repl):
    def __init__(self, connection, address, **repl_kwargs):
        self.connection = connection
        self.address = address

        super().__init__(**repl_kwargs)

    def send(self, data):
        logger.debug(
            'sending: %s',
            repr(data),
        )

        self.connection.send(data)

    def write_message(self, message):
        exit_code, payload = message

        if not exit_code:
            return

        self.send(payload)

    def setup(self):
        self.write_message(encode_set_message('banner', self.banner))
        self.write_message(encode_set_message('warnings', self.warnings))
        self.write_message(encode_set_message('ps1', self.ps1))
        self.write_message(encode_set_message('ps2', self.ps2))
        self.write_message(encode_ready_message())

    def write(self, string):
        self.write_message(encode_write_message(string))

    def run(self, command):
        exit_code = super().run(command)

        self.write_message(encode_exit_code_message(exit_code))


class ReplConnection:
    def __init__(self, repl_server, connection, address, repl_domain,
                 **repl_kwargs):

        self.repl_server = repl_server
        self.connection = connection
        self.address = address
        self.repl_domain = repl_domain

        self.repl = SocketRepl(
            connection=self.connection,
            address=self.address,
            **repl_kwargs,
        )

        self.repl.set_domain(self.repl_domain)

    def shutdown(self):
        self.repl.shutdown()

    def handle_message(self, raw_message):
        message_is_valid, message_type, payload = decode_message(raw_message)

        logger.debug(
            '%s message received: %s',
            'valid' if message_is_valid else 'invalid',
            raw_message,
        )

        if not message_is_valid:
            return

        # ping (empty line)
        if message_type == MESSAGE_TYPE.PING:
            if self.repl.variables['repeat_last_command_on_enter']:
                command = self.repl.history[-1]

                self.repl.write(command + '\n')
                self.repl.run(command)

            self.repl.write_message(encode_pong_message())

        # run
        elif message_type == MESSAGE_TYPE.RUN:
            self.repl.run(payload)

        # completion request
        elif message_type == MESSAGE_TYPE.COMPLETION_REQUEST:
            text, state, line_buffer = payload

            response = self.repl.completer.complete(text, state, line_buffer)

            self.repl.write_message(
                encode_completion_response_message(response),
            )

    def run_single_threaded(self):
        message_buffer = bytes()

        while True:
            data = self.connection.recv(1)

            if not data:
                break

            message_buffer += data

            if message_buffer[-1] == END_OF_TRANSMISSION[0]:
                self.handle_message(message_buffer)
                message_buffer = bytes()

                continue



import threading
import socket
import os

from rlpython.repl_connection import ReplConnection
from rlpython.utils.url import parse_url
from rlpython.repl import Repl


class ReplServer:
    def __init__(self, url, permissions, repl_domain=Repl.DOMAIN.NETWORK,
                 print=print, **repl_kwargs):

        self.url = url
        self.permissions = permissions
        self.repl_domain = repl_domain
        self._print = print
        self.repl_kwargs = repl_kwargs

        self.scheme, self.host, self.port = parse_url(url)

    def setup(self):
        # network
        if self.scheme == 'rlpython':
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            bind_arg = (self.host, self.port)

        # unix domain socket
        elif self.scheme == 'file':
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            bind_arg = self.host

            try:
                os.remove(self.host)

            except OSError:
                pass

        self.sock.bind(bind_arg)

        if self.scheme == 'file':
            # set file permmisions
            os.chmod(self.host, int(self.permissions, 8))

        self.sock.listen(1)

    def print_bind_informations(self):
        if not self._print:
            return

        if self.scheme == 'rlpython':
            sock_name = self.sock.getsockname()

            self._print(
                'rlpython: running on {}:{}'.format(
                    sock_name[0],
                    sock_name[1],
                )
            )

        elif self.scheme == 'file':
            self._print(
                'rlpython: running on file://{}'.format(self.host),
            )

    def shutdown(self):
        self._running = False

        # remove unix domain socket
        if self.scheme == 'file':
            try:
                os.remove(self.host)

            except OSError:
                pass

    def __enter__(self):
        self.setup()
        self.print_bind_informations()
        self.run_multi_session(**self.repl_kwargs)

        return self

    def __exit__(self, type, value, traceback):
        self.shutdown()

    def get_host(self):
        return self.sock.getsockname()[0]

    def get_port(self):
        return self.sock.getsockname()[1]

    def run_single_session(self, **repl_kwargs):
        repl_connection = None

        repl_kwargs = {
            **self.repl_kwargs,
            **repl_kwargs,
        }

        try:
            connection, address = self.sock.accept()

            repl_connection = ReplConnection(
                repl_server=self,
                connection=connection,
                address=address,
                repl_domain=self.repl_domain,
                **repl_kwargs,
            )

            repl_connection.run_single_threaded()

        except KeyboardInterrupt:
            return

        finally:
            if repl_connection:
                repl_connection.shutdown()

    def run_multi_session(self, **repl_kwargs):
        def _server_thread():
            while self._running:
                try:
                    connection, address = self.sock.accept()

                except(socket.timeout, OSError):
                    continue

                connection_count[0] += 1

                repl_connection = ReplConnection(
                    repl_server=self,
                    connection=connection,
                    address=address,
                    repl_domain=self.repl_domain,
                    **repl_kwargs,
                )

                threading.Thread(
                    name='rlpython REPL Session {}'.format(
                        connection_count[0],
                    ),
                    daemon=True,
                    target=repl_connection.run_single_threaded,
                ).start()

        repl_kwargs = {
            **self.repl_kwargs,
            **repl_kwargs,
        }

        self.sock.settimeout(1)
        self._running = True

        connection_count = [0]

        threading.Thread(
            name='rlpython REPL Server',
            daemon=True,
            target=_server_thread,
        ).start()



EXPR_START = '{{'
EXPR_STOP = '}}'
EXPR_START_LEN = len(EXPR_START)
EXPR_STOP_LEN = len(EXPR_STOP)


class TemplatingEngine:
    def __init__(self, repl):
        self.repl = repl

    def split(self, string):
        strings = []

        cursor = 0
        recording = False

        while cursor <= len(string):
            if recording:

                # searching for EXPR_STOP
                if string[cursor:cursor+EXPR_STOP_LEN].startswith(EXPR_STOP):
                    strings.append(
                        (True, string[EXPR_START_LEN:cursor],)
                    )

                    string = string[cursor+EXPR_STOP_LEN:]
                    cursor = 0
                    recording = False

                    continue

            else:

                # searching for EXPR_START
                if string[cursor:].startswith(EXPR_START):
                    recording = True

                    if cursor > 0:
                        strings.append(
                            (False, string[:cursor],)
                        )

                        string = string[cursor:]
                        cursor = 0

                    continue

            cursor += 1

        if not len(string) == 0:
            strings.append(
                (False, string,)
            )

        return strings

    def render(self, template):
        output = ''
        parts = self.split(template)

        for is_expression, string in parts:
            if is_expression:
                resolved_expression = self.repl.python_runtime.eval(
                    string,
                    safe=False,
                )

                if resolved_expression is None:
                    continue

                string = str(resolved_expression)

            output += string

        return output

    def is_template(self, string):
        return EXPR_START in string



from copy import deepcopy

from rlpython.utils.strings import get_length


def write_table(rows, write):
    rows = deepcopy(rows)

    # find column widths
    column_widths = [[] for i in range(len(rows[0]))]

    for row_index, row in enumerate(rows):
        for column_index, column in enumerate(row):
            lines = str(column).strip().splitlines()
            rows[row_index][column_index] = lines

            column_widths[column_index].append(
                max([0] + [get_length(line) for line in lines])
            )

    for index, column_width in enumerate(column_widths):
        column_widths[index] = max(column_widths[index])

    # write table out
    def get_row_len(row):
        return max([len(i) for i in row])

    def write_divider():
        for column_width in column_widths:
            write('+')
            write('-' * (column_width + 2))

        write('+\n')

    def write_empty_row():
        for column_width in column_widths:
            write('|')
            write(' ' * (column_width + 2))

        write('|\n')

    def write_row(row):
        lines = get_row_len(row)
        row_index = rows.index(row)

        if lines > 1 and row_index > 1:
            write_empty_row()

        for line_index in range(lines):
            write('| ')

            for column_index, column in enumerate(row):
                if line_index < len(column):
                    line = column[line_index]

                else:
                    line = ''

                write(line.ljust(column_widths[column_index]))
                write(' | ')

            write('\n')

        if(lines > 1 and
           row_index < len(rows) - 1 and
           get_row_len(rows[row_index+1]) < 2):

            write_empty_row()

    write_divider()
    write_row(rows[0])
    write_divider()

    for row in rows[1:]:
        write_row(row)

    write_divider()



from rlpython.utils.argument_parser import ReplArgumentParser
from rlpython.utils.table import write_table


class AliasesCommand:
    """
    Manage REPL aliases
    """

    NAME = 'alias'

    def __init__(self, repl):
        self.repl = repl

    def run(self, argv):
        # parse command line
        argument_parser = ReplArgumentParser(repl=self.repl, prog='alias')
        argument_parser.add_argument('name', nargs='?')

        arguments = argument_parser.parse_args(argv[1:])

        # print all variables
        if not arguments.name:
            table = [
                ['Name', 'Value'],
            ]

            for name, value in self.repl.aliases.aliases.items():
                table.append([
                    name,
                    value,
                ])

            write_table(table, self.repl.write)

            return 0

        # print variable
        if '=' not in arguments.name:
            self.repl.write(
                self.repl.aliases.aliases.get(arguments.name, ' ') + '\n')

            return 0

        # set variable
        name, value = arguments.name.split('=', 1)
        self.repl.aliases.aliases[name] = value

        return 0



COLOR_CODES = {
    'black': '30',
    'red': '31',
    'green': '32',
    'yellow': '33',
    'blue': '34',
    'magenta': '35',
    'cyan': '36',
    'white': '37',
}

STYLE_CODES = {
    'normal': '0',
    'bright': '1',
    'underlined': '2',
    'negative': '3',
}


def color(string, fg='', bg='', style='normal', reset=True):
    if fg:
        fg = COLOR_CODES[fg]

    if style:
        style = STYLE_CODES[style]

    string = '\001\033[{};{}m\002{}'.format(style, fg, string)

    if reset:
        string += '\001\033[0m\002'

    return string


def get_length(string):
    length = 0
    quoted = False

    for character in string:
        if character == '\001':
            quoted = True

        elif character == '\002':
            quoted = False

        elif not quoted:
            length += 1

    return length



from asyncio.base_events import BaseEventLoop
import asyncio
import sys

from rlpython.utils.gc_utils import get_objects_by_class


def get_all_tasks(loop):
    if sys.version_info >= (3, 7):
        return asyncio.all_tasks(loop=loop)

    return asyncio.Task.all_tasks(loop=loop)


def get_all_loops():
    return get_objects_by_class(object_class=BaseEventLoop)



import gc


def get_objects_by_class(object_class):
    objects = []

    for obj in gc.get_objects():
        if isinstance(obj, object_class):
            objects.append(obj)

    return objects


def get_object_by_id(object_id):
    object_id = str(object_id)

    for obj in gc.get_objects():
        if str(id(obj)) == object_id:
            return obj



import inspect
import logging
import os


def embed(single_threaded=False, bind='', permissions='600',
          multi_session=False, started_from_cmd_line=False, print=print,
          debug=False, **repl_kwargs):

    from rlpython.frontend import start_frontend
    from rlpython.repl_server import ReplServer
    from rlpython.repl import Repl

    # debug mode
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        os.environ['RLPYTHON_DEBUG'] = 'True'

    # use namespace of caller instead of own if nothing is set
    if 'globals' not in repl_kwargs and 'locals' not in repl_kwargs:
        stack = inspect.stack()
        frame_info = stack[1]

        repl_kwargs['globals'] = frame_info.frame.f_globals
        repl_kwargs['locals'] = frame_info.frame.f_locals

    # setup warnings
    if 'warnings' not in repl_kwargs:
        repl_kwargs['warnings'] = []

    if not started_from_cmd_line:
        repl_kwargs['warnings'].append('running single threaded: cancellation using CTRL-C will not work')  # NOQA

    if single_threaded and not bind and not started_from_cmd_line:
        repl_kwargs['warnings'].append('running single threaded: Use "!" to cancel multi line statements')  # NOQA

    # network embed
    if bind:
        single_threaded = True  # FIXME

        # single threaded
        if single_threaded:
            repl_server = ReplServer(
                url=bind,
                permissions=permissions,
                repl_domain=Repl.DOMAIN.NETWORK,
                print=print,
                **repl_kwargs,
            )

            if multi_session:
                return repl_server

            try:
                repl_server.setup()
                repl_server.print_bind_informations()
                repl_server.run_single_session(**repl_kwargs)

            except OSError as exception:
                exit('rlpython: ERROR: {}'.format(exception.args[1]))

            finally:
                repl_server.shutdown()

        # multi threaded
        else:
            raise NotImplementedError

    # local embed
    else:

        # single threaded
        if single_threaded:
            repl = Repl(**repl_kwargs)

            try:
                repl.interact()

            finally:
                repl.shutdown()

        # multi threaded
        else:
            repl_server = ReplServer(
                url='localhost:0',
                permissions=permissions,
                repl_domain=Repl.DOMAIN.LOCAL_NETWORK,
            )

            repl_server.setup()
            port = repl_server.get_port()

            start_frontend(port)

            repl_server.run_single_session(**repl_kwargs)



import subprocess
import shutil
import shlex
import os

EDITORS = [
    'vim',
    'vi',
    'nano',
]


class NoEditorError(Exception):
    pass


def run_editor(filename, lineno=None):
    editor = os.environ.get('EDITOR')

    if not editor:
        for name in EDITORS:
            if shutil.which(name):
                editor = name

                break

    if not editor:
        raise NoEditorError()

    command = '{} {}'.format(editor, filename)

    if lineno is not None:
        command = '{} +{}'.format(command, lineno)

    subprocess.call(shlex.split(command))



from traceback import format_exception
import readline
import json
import sys
import os

from rlpython.runtimes.command_runtime import CommandRuntime
from rlpython.runtimes.python_runtime import PythonRuntime
from rlpython.runtimes.shell_runtime import ShellRuntime
from rlpython.utils.gc_utils import get_object_by_id
from rlpython.templating import TemplatingEngine
from rlpython.completion import Completer
from rlpython.utils.strings import color
from rlpython.aliases import Aliases
from rlpython import VERSION_STRING

DEFAULT_HISTORY_FILE = '~/.rlpython.history'
DEFAULT_HISTORY_SIZE = 1000
DEFAULT_PROMPT_PREFIX = ''
DEFAULT_PROMPT = '>>> '
DEFAULT_PROMPT_PS2 = '... '

DEFAULT_BANNER = """Python {} on {}
rlpython {}
Type '?' for help
""".format(
    sys.version,
    sys.platform,
    VERSION_STRING,
)

DEFAULT_VARIABLES = {
    'pretty_print': True,
    'repeat_last_command_on_enter': False,
}

HELP_TEXT = """
rlpython
========

rlpython is a GNU Readline based Python REPL, fully compatible with
Python's syntax.


Help / Inspection
-----------------

?             Show this help text
object?       Inspect given object
object??      Show help text of given object
%edit object  Open the source code of given object if available
help          Python's builtin help system


REPL Commands
-------------

{commands}
"""


class Repl:
    class DOMAIN:
        LOCAL = 1
        NETWORK = 2
        LOCAL_NETWORK = 3

    def __init__(self, banner=DEFAULT_BANNER, warnings=[],
                 prompt_prefix=DEFAULT_PROMPT_PREFIX,
                 prompt=DEFAULT_PROMPT,
                 prompt_ps2=DEFAULT_PROMPT_PS2,
                 history_file=DEFAULT_HISTORY_FILE,
                 history_size=DEFAULT_HISTORY_SIZE,
                 globals={}, locals={}, commands=[], **variables):

        self.banner = banner
        self.warnings = warnings
        self.prompt_prefix = prompt_prefix
        self.ps1 = prompt
        self.ps2 = prompt_ps2
        self.history_file = os.path.expanduser(history_file)
        self.history_size = history_size
        self.globals = globals
        self.locals = locals

        self.exit_code = 0

        self.variables = {
            **DEFAULT_VARIABLES,
            **variables,
        }

        # setup templating engine
        self.templating_engine = TemplatingEngine(repl=self)

        # setup completion
        self.completer = Completer(repl=self)

        # setup readline
        readline.set_auto_history(False)
        readline.parse_and_bind('tab: complete')
        readline.set_completer(self.complete)
        readline.set_completer_delims(' \t\n`~!@#$%^&*()-=+[{]}\\|;:\'",<>?')

        self.read_history()

        # setup runtimes
        self.python_runtime = PythonRuntime(
            repl=self,
            globals=globals,
            locals=locals,
        )

        self.command_runtime = CommandRuntime(
            repl=self,
        )

        for command in commands:
            self.install_command(command)

        self.shell_runtime = ShellRuntime(
            repl=self,
        )

        # setup aliasses
        self.aliases = Aliases(repl=self)

        # finish
        self.clear_line_buffer()
        self.setup()

    def shutdown(self):
        self.python_runtime.shutdown()
        self.command_runtime.shutdown()

        if self.domain == self.DOMAIN.LOCAL:
            self.write_history()

    # domain ##################################################################
    @property
    def domain(self):
        if not hasattr(self, '_domain'):
            self._domain = self.DOMAIN.LOCAL

        return self._domain

    def set_domain(self, domain):
        self._domain = domain

    # utils ###################################################################
    def get_object_by_id(self, object_id):
        return get_object_by_id(object_id)

    # commands ################################################################
    def install_command(self, command):
        self.command_runtime.install_command(command)

    # history #################################################################
    def read_history(self):
        self.history = []
        readline.clear_history()

        try:
            for line in open(self.history_file, 'r'):
                item = json.loads(line)
                self.add_history(item)

        except Exception:
            pass

    def add_history(self, item):
        # skip duplicates
        if self.history and item == self.history[-1]:
            return

        self.history.append(item)

        # rotate history
        self.history = self.history[self.history_size*-1:]

        for item in self.history:
            readline.add_history(item)

    def write_history(self):
        try:
            with open(self.history_file, 'w+') as f:
                for item in self.history:
                    f.write(json.dumps(item) + '\n')

        except Exception:
            pass

    # error management ########################################################
    def write_exception(self, exception, prefix=''):
        lines = format_exception(
            etype=type(exception),
            value=exception,
            tb=exception.__traceback__,
        )

        text = ''.join(lines)

        if prefix:
            lines = text.splitlines()

            for index, line in enumerate(lines):
                lines[index] = prefix + line

            text = '\n'.join(lines) + '\n'

        self.write(color(text, fg='red'))

    def write_warning(self, string):
        self.write(color('WARNING: {}'.format(string), fg='yellow'))

    def write_error(self, string):
        self.write(color('ERROR: {}'.format(string), fg='red'))

    # line buffer #############################################################
    def clear_line_buffer(self):
        self.line_buffer = ''

    def validate_line_buffer(self):
        try:
            line_buffer = self.aliases.resolve(self.line_buffer)

            if line_buffer.startswith('%'):
                return self.command_runtime.validate_source(line_buffer)

            elif line_buffer.startswith('!'):
                return self.shell_runtime.validate_source(line_buffer)

            return self.python_runtime.validate_source(line_buffer)

        except Exception as exception:
            self.exit_code = 1

            self.write_exception(exception, prefix='rlpython: ')

            return False

    # I/O #####################################################################
    def write_warnings(self):
        for warning in self.warnings:
            self.write(color('WARNING: {}\n'.format(warning), fg='yellow'))

    def write_banner(self):
        self.write(self.banner)

    def write_help(self):
        text = HELP_TEXT.format(
            commands=self.command_runtime.gen_help_text(),
        )

        self.write(text.strip() + '\n')

    def gen_prompt(self):
        prompt_color = 'green'

        if self.exit_code != 0:
            prompt_color = 'red'

        ps1 = color(
            self.prompt_prefix + self.ps1,
            fg=prompt_color,
            style='bright',
        )

        ps2 = color(
            self.prompt_prefix + self.ps2,
            fg=prompt_color,
            style='normal',
        )

        if not self.line_buffer:
            return ps1

        if self.validate_line_buffer():
            return ps1

        return ps2

    def interact(self):
        while True:
            try:
                self.line_buffer += input(self.gen_prompt()) + '\n'

                # handle empty lines
                if not self.line_buffer.strip():
                    self.handle_empty_line()

                    continue

            except KeyboardInterrupt:
                self.handle_ctrl_c()

                continue

            except EOFError:
                if self.handle_ctrl_d():
                    return

                continue

            # emulated CTRL-C
            if self.line_buffer.strip().endswith('!'):
                self.handle_ctrl_c()

                continue

            # multi line
            if not self.validate_line_buffer():
                continue

            self.add_history(self.line_buffer.strip())
            self.run(self.line_buffer)
            self.clear_line_buffer()

    # application code ########################################################
    def write(self, string):
        print(string, end='')

    def setup(self):
        self.write_banner()
        self.write_warnings()

    def complete(self, text, state):
        line_buffer = readline.get_line_buffer()

        return self.completer.complete(text, state, line_buffer)

    def handle_empty_line(self):
        self.clear_line_buffer()

        if self.variables['repeat_last_command_on_enter']:
            command = self.history[-1]

            self.write(command + '\n')
            self.run(command)

    def handle_ctrl_d(self):
        self.clear_line_buffer()
        self.write('\n')

        while True:
            try:
                answer = input('Do you really want to exit ([y]/n)? ')
                answer = answer.strip().lower()

                if answer in ('y', ''):
                    return True

                if answer == 'n':
                    return False

            except(EOFError, KeyboardInterrupt):
                self.write('\n')

                return False

    def handle_ctrl_c(self):
        self.clear_line_buffer()
        self.write('^C\n')
        self.exit_code = 1

    def run(self, command):
        try:
            # aliasses
            command = self.aliases.resolve(command)

            # templating
            if self.templating_engine.is_template(command):
                command = self.templating_engine.render(command)
                self.write(command)

            # help
            if command.strip() == '?':
                self.write_help()
                self.exit_code = 0

            # repl commands
            elif command.startswith('%'):
                self.exit_code = self.command_runtime.run(command)

            # shell commands
            elif command.startswith('!'):
                self.exit_code = self.shell_runtime.run(command)

            # python code
            else:
                self.exit_code = self.python_runtime.run(command)

        except Exception as exception:
            self.exit_code = 1

            self.write_exception(exception, prefix='rlpython: ')

        return self.exit_code
